/* mtx2wav.c - Convert a standard (BASIC) MTX file to WAV format */

#include <stdint.h>
#include <stdio.h>

int nMTXBytesOut = 0;
uint8_t buffer[16384];

FILE * create_wav (const char *psFile)
    {
    int nRateOut = 4800;             /* WAV output file sample rate */
    int iVal;
    uint16_t wVal;
    FILE *pf = fopen (psFile, "wb");
    if ( pf == NULL ) return pf;
    fwrite ("RIFF", sizeof (char), 4, pf);
    fwrite (&nMTXBytesOut, sizeof (nMTXBytesOut), 1, pf);
    fwrite ("WAVE", sizeof (char), 4, pf);
    fwrite ("fmt ", sizeof (char), 4, pf);
    iVal = 18;  /* Chunk size */
    fwrite (&iVal, sizeof (iVal), 1, pf);
    wVal = 1;   /* Format */
    fwrite (&wVal, sizeof (wVal), 1, pf);
    wVal = 1;   /* Number of channels */
    fwrite (&wVal, sizeof (wVal), 1, pf);
    fwrite (&nRateOut, sizeof (nRateOut), 1, pf);
    fwrite (&nRateOut, sizeof (nRateOut), 1, pf);
    wVal = 1;   /* Block align */
    fwrite (&wVal, sizeof (wVal), 1, pf);
    iVal = 8;   /* Bits per sample */
    fwrite (&iVal, sizeof (iVal), 1, pf);
    fwrite ("data", sizeof (char), 4, pf);
    fwrite (&nMTXBytesOut, sizeof (nMTXBytesOut), 1, pf);
    return pf;
    }

void close_wav (FILE *pf)
    {
    /* Output final bit */
    uint8_t byOut = 255;
    fwrite (&byOut, sizeof (byOut), 1, pf);
    byOut = 0;
    fwrite (&byOut, sizeof (byOut), 1, pf);
    nMTXBytesOut += 2;
    /* Size of data chunk */
    fseek (pf, 42, SEEK_SET);
    fwrite (&nMTXBytesOut, sizeof (nMTXBytesOut), 1, pf);
    /* Size of WAVE file */
    nMTXBytesOut += 34;
    fseek (pf, 4, SEEK_SET);
    fwrite (&nMTXBytesOut, sizeof (nMTXBytesOut), 1, pf);
    fseek (pf, 0, SEEK_END);
    fclose (pf);
    }

void write_block (FILE *pf, const uint8_t *pb, int nLen)
    {
    // Write lead-in

    uint8_t byOut;
    for (int i = 0; i < 1500; ++i)
        {
        byOut = 255;
        fwrite (&byOut, sizeof (byOut), 1, pf);
        byOut = 0;
        fwrite (&byOut, sizeof (byOut), 1, pf);
        nMTXBytesOut += 2;
        }

    // Write start of block marker
    
    byOut = 255;
    fwrite (&byOut, sizeof (byOut), 1, pf);
    byOut = 0;
    fwrite (&byOut, sizeof (byOut), 1, pf);
    fwrite (&byOut, sizeof (byOut), 1, pf);
    fwrite (&byOut, sizeof (byOut), 1, pf);
    nMTXBytesOut += 4;

    // Write data

    for (int i = 0; i < nLen; ++i)
        {
        uint8_t byData = *pb;
        for (int j = 0; j < 8; ++j)
            {
            if ( byData & 0x01 )
                {
                byOut = 255;
                fwrite (&byOut, sizeof (byOut), 1, pf);
                fwrite (&byOut, sizeof (byOut), 1, pf);
                byOut = 0;
                fwrite (&byOut, sizeof (byOut), 1, pf);
                fwrite (&byOut, sizeof (byOut), 1, pf);
                nMTXBytesOut += 4;
                }
            else
                {
                byOut = 255;
                fwrite (&byOut, sizeof (byOut), 1, pf);
                byOut = 0;
                fwrite (&byOut, sizeof (byOut), 1, pf);
                nMTXBytesOut += 2;
                }
            byData >>= 1;
            }
        ++pb;
        }
    }

int main (int nArg, const char *psArg[])
    {
    static const char *psBlkErr = "Block %d: Expecting %d bytes, read %d bytes.\n";
    
    // Optionally display usage
    
    if ( nArg != 3 )
        {
        fprintf (stderr, "Usage: %s <MTX file> <WAV file>\n", psArg[0]);
        return 2;
        }

    // Open files
    
    FILE *pfMTX = fopen (psArg[1], "rb");
    if ( pfMTX == NULL )
        {
        fprintf (stderr, "Unable to open input file %s\n", psArg[1]);
        return 3;
        }
    FILE *pfWAV = create_wav (psArg[2]);
    if ( pfWAV == NULL )
        {
        fprintf (stderr, "Unable to open output file %s\n", psArg[2]);
        return 4;
        }

    // Name block

    printf ("Name block\n");
    int nBlk = 1;
    int nRead = fread (buffer, 1, 18, pfMTX);
    if ( nRead != 18 )
        {
        printf (psBlkErr, nBlk, 18, nRead);
        return 5;
        }
    if ( buffer[0] != 0xFF )
        {
        fprintf (stderr, "Name block does not begin with 0xFF\n");
        return 5;
        }
    printf ("File name: ");
    for (int i = 1; i < 16; ++i)
        {
        if ((buffer[i] >= 0x20) && (buffer[i] < 0x7F)) printf ("%c", buffer[i]);
        else printf ("\\%02X", buffer[i]);
        }
    int iStkLim = (buffer[17] << 8) | buffer[16];
    printf ("\nSTKLIM = 0x%04X\n", iStkLim);
    if ( iStkLim > 0xFA7A )
        {
        fprintf (stderr, "Invalid STKLIM value: 0x%04X\n", iStkLim);
        return 5;
        }
    write_block (pfWAV, buffer, 18);

    // System variables

    printf ("System Variables Block\n");
    ++nBlk;
    int nLen = 0xFB4B - iStkLim;
    nRead = fread (buffer, 1, nLen, pfMTX);
    if ( nRead != nLen )
        {
        printf (psBlkErr, nBlk, nLen, nRead);
        return 5;
        }
    int iLstPg = buffer[0xFA7A - iStkLim];
    int iVarNam = (buffer[0xFA7C - iStkLim] << 8) | buffer[0xFA7B - iStkLim];
    int iCalcSt = (buffer[0xFA82 - iStkLim] << 8) | buffer[0xFA81 - iStkLim];
    int iArrTop = (buffer[0xFACE - iStkLim] << 16 ) | (buffer[0xFACD - iStkLim] << 8) | buffer[0xFACC - iStkLim];
    printf ("LSTPAG = %d\nVARNAM = 0x%04X\nCALCST = 0x%04X\nARRTOP = 0x%06X\n",
        iLstPg, iVarNam, iCalcSt, iArrTop);
    if ( iCalcSt < iVarNam )
        {
        fprintf (stderr, "Invalid CALCST (0x%04X) < VARNAM (0x%04X)\n", iCalcSt, iVarNam);
        return 5;
        }
    write_block (pfWAV, buffer, nLen);

    // Program

    while ( iArrTop > 0 )
        {
        ++nBlk;
        nLen = iArrTop;
        if ( nLen > 0x4000 ) nLen = 0x4000;
        printf ("Program Block %d: Length = %d bytes\n", nBlk - 2, nLen);
        nRead = fread (buffer, 1, nLen, pfMTX);
        if ( nRead != nLen )
            {
            printf (psBlkErr, nBlk, nLen, nRead);
            return 5;
            }
        write_block (pfWAV, buffer, nLen);
        iArrTop -= nLen;
        }

    // Program variables

    if ( iCalcSt > iVarNam )
        {
        ++nBlk;
        nLen = iCalcSt - iVarNam;
        printf ("Program Variables Block: Length = %d bytes\n", nLen);
        nRead = fread (buffer, 1, nLen, pfMTX);
        if ( nRead != nLen )
            {
            printf (psBlkErr, nBlk, nLen, nRead);
            return 5;
            }
        write_block (pfWAV, buffer, nLen);
        }

    // Close WAV file

    close_wav (pfWAV);

    // Test for remaining data

    long nPos = ftell (pfMTX);
    fseek (pfMTX, 0L, SEEK_END);
    nPos = ftell (pfMTX) - nPos;
    if ( nPos > 0 )
        {
        printf ("Warning: %ld bytes at end of MTX file ignored\n", nPos);
        return 1;
        }

    return 0;
    }
